#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .base import ModifierScoreFn, ScoreOps as so


class Normalizer1(ModifierScoreFn):
    """Do normalizing: score = 1 / (1 + sqrt(score))"""

    def __init__(self):
        super().__init__(modifier='reciprocal1p')

    def __call__(self, last_score, *args, **kwargs):
        return super().__call__(so.sqrt(last_score))


class Normalizer2(ModifierScoreFn):
    """Do normalizing: score = 1 / (1 + score / num_dim)"""

    def __init__(self, num_dim: int):
        super().__init__(modifier='reciprocal1p', factor=1.0 / num_dim, factor_name='1/num_dim')


class Normalizer3(Normalizer2):
    """Do normalizing: score = 1 / (1 + sqrt(score) / num_dim)"""

    def __call__(self, last_score, *args, **kwargs):
        return super().__call__(so.sqrt(last_score))


class Normalizer4(ModifierScoreFn):
    """Do normalizing: score = 1 - score / num_bytes """

    def __init__(self, num_bytes: int):
        super().__init__(modifier='invert1p', factor=1.0 / num_bytes, factor_name='1/num_bytes')


class Normalizer5(ModifierScoreFn):
    """Do normalizing: score = 1 / (1 + sqrt(abs(score)))"""

    def __init__(self):
        super().__init__(modifier='reciprocal1p')

    def __call__(self, last_score, *args, **kwargs):
        return super().__call__(so.sqrt(so.abs(last_score)))



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List
import numpy as np
from ..base import BaseVideoEncoder
from ...helper import batching, get_first_available_gpu


class YouTube8MEncoder(BaseVideoEncoder):
    batch_size = 64

    def __init__(self, model_dir: str,
                 model_name: str,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_dir = model_dir
        self.model_name = model_name
        self.max_num_frames = 300
        self.embedding_dim = 1152

    def post_init(self):
        import tensorflow as tf
        import os
        os.environ['CUDA_VISIBLE_DEVICES'] = str(get_first_available_gpu())

        g = tf.Graph()
        with g.as_default():
            checkpoint_file = os.path.join(self.model_dir, self.model_name,
                                           "inference_model")
            meta_graph_location = checkpoint_file + ".meta"
            saver = tf.train.import_meta_graph(meta_graph_location, clear_devices=True)

            config = tf.ConfigProto(log_device_placement=False)
            if self.on_gpu:
                config.gpu_options.allow_growth = True
            self.sess = tf.Session(config=config)
            saver.restore(self.sess, checkpoint_file)

            self.input_tensor = tf.get_collection("input_batch_raw")[0]
            self.num_frames_tensor = tf.get_collection("num_frames")[0]
            self.predictions_tensor = self.sess.graph.get_tensor_by_name("tower/gates/MatMul:0")

    def encode(self, data: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        def _padding(data):
            _data = np.array([np.concatenate((d, np.zeros((self.max_num_frames - d.shape[0], self.embedding_dim), dtype=np.float32)),
                            axis=0) if d.shape[0] < self.max_num_frames else d[:self.max_num_frames] for d in data])
            return _data.reshape((-1, self.max_num_frames, self.embedding_dim))

        @batching
        def _encode(_, data):
            num_frames_list = list(map(lambda x: x.shape[0], data))
            predictions_val, = self.sess.run([self.predictions_tensor],
                        feed_dict={self.input_tensor: _padding(data),
                                   self.num_frames_tensor: np.array(num_frames_list)})
            return np.array(predictions_val).astype(np.float32)

        return _encode(self, _padding(data))





import json
import unittest
from pprint import pprint

from gnes.proto import gnes_pb2
from gnes.score_fn.base import get_unary_score, CombinedScoreFn, ModifierScoreFn
from gnes.score_fn.chunk import WeightedChunkScoreFn
from gnes.score_fn.normalize import Normalizer1, Normalizer2, Normalizer3, Normalizer4


class TestScoreFn(unittest.TestCase):
    def test_basic(self):
        a = get_unary_score(0.5)
        b = get_unary_score(0.7)
        print(a)
        print(b.explained)

    def test_op(self):
        a = get_unary_score(0.5)
        b = get_unary_score(0.7)
        sum_op = CombinedScoreFn(score_mode='sum')
        c = sum_op(a, b)
        self.assertAlmostEqual(c.value, 1.2)

        sq_op = ModifierScoreFn(modifier='square')
        c = sum_op(a, sq_op(b))
        self.assertAlmostEqual(c.value, 0.99)
        print(c)

    def test_normalizer(self):
        a = get_unary_score(0.5)
        norm_op = Normalizer1()
        b = norm_op(a)
        pprint(json.loads(b.explained))

        a = get_unary_score(0.5)
        norm_op = Normalizer2(2)
        b = norm_op(a)
        pprint(json.loads(b.explained))
        self.assertAlmostEqual(b.value, 0.8)

        a = get_unary_score(0.5)
        norm_op = Normalizer3(2)
        b = norm_op(a)
        pprint(json.loads(b.explained))
        self.assertAlmostEqual(b.value, 0.7387961283389092)

        a = get_unary_score(0.5)
        norm_op = Normalizer4(2)
        b = norm_op(a)
        pprint(json.loads(b.explained))
        self.assertEqual(b.value, 0.75)

        norm_op = ModifierScoreFn('none')
        b = norm_op(a)
        pprint(json.loads(b.explained))
        self.assertEqual(b.value, 0.5)

        q_chunk = gnes_pb2.Chunk()
        q_chunk.weight = 0.5
        q_chunk.offset = 1
        d_chunk = gnes_pb2.Chunk()
        d_chunk.weight = 0.7
        d_chunk.offset = 2
        rel_score = get_unary_score(2)
        _op = WeightedChunkScoreFn()
        c = _op(rel_score, q_chunk, d_chunk)
        pprint(json.loads(c.explained))
        self.assertAlmostEqual(c.value, 0.7)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List, Any, Union, Callable, Tuple

import numpy as np

from ..base import TrainableBase, CompositionalTrainableBase
from ..proto import gnes_pb2, blob2array
from ..score_fn.base import get_unary_score, ModifierScoreFn


class BaseIndexer(TrainableBase):
    def __init__(self,
                 normalize_fn: 'BaseScoreFn' = ModifierScoreFn(),
                 score_fn: 'BaseScoreFn' = ModifierScoreFn(),
                 is_big_score_similar: bool = False,
                 *args, **kwargs):
        """
        Base indexer, a valid indexer must implement `add` and `query` methods
        :type score_fn: advanced score function
        :type normalize_fn: normalizing score function
        :type is_big_score_similar: when set to true, then larger score means more similar
        """
        super().__init__(*args, **kwargs)
        self.normalize_fn = normalize_fn
        self.score_fn = score_fn
        self.is_big_score_similar = is_big_score_similar

    def add(self, keys: Any, docs: Any, weights: List[float], *args, **kwargs):
        pass

    def query(self, keys: Any, *args, **kwargs) -> List[Any]:
        pass

    def query_and_score(self, q_chunks: List[Union['gnes_pb2.Chunk', 'gnes_pb2.Document']], top_k: int) -> List[
        'gnes_pb2.Response.QueryResponse.ScoredResult']:
        raise NotImplementedError


class BaseChunkIndexer(BaseIndexer):

    def add(self, keys: List[Tuple[int, int]], vectors: np.ndarray, weights: List[float], *args, **kwargs):
        pass

    def query(self, keys: np.ndarray, top_k: int, *args, **kwargs) -> List[List[Tuple]]:
        pass

    def query_and_score(self, q_chunks: List['gnes_pb2.Chunk'], top_k: int, *args, **kwargs) -> List[
        'gnes_pb2.Response.QueryResponse.ScoredResult']:
        vecs = [blob2array(c.embedding) for c in q_chunks]
        queried_results = self.query(np.stack(vecs), top_k=top_k)
        results = []
        for q_chunk, topk_chunks in zip(q_chunks, queried_results):
            for _doc_id, _offset, _weight, _relevance in topk_chunks:
                r = gnes_pb2.Response.QueryResponse.ScoredResult()
                r.chunk.doc_id = _doc_id
                r.chunk.offset = _offset
                r.chunk.weight = _weight
                _score = get_unary_score(value=_relevance,
                                         name=self.__class__.__name__,
                                         operands=[
                                             dict(name='doc_chunk',
                                                  doc_id=_doc_id,
                                                  offset=_offset),
                                             dict(name='query_chunk',
                                                  offset=q_chunk.offset)])
                _score = self.normalize_fn(_score)
                _score = self.score_fn(_score, q_chunk, r.chunk)
                r.score.CopyFrom(_score)
                results.append(r)
        return results


class BaseDocIndexer(BaseIndexer):

    def add(self, keys: List[int], docs: Any, weights: List[float], *args, **kwargs):
        pass

    def query(self, keys: List[int], *args, **kwargs) -> List['gnes_pb2.Document']:
        pass

    def query_and_score(self, docs: List['gnes_pb2.Response.QueryResponse.ScoredResult'], *args, **kwargs) -> List[
        'gnes_pb2.Response.QueryResponse.ScoredResult']:
        keys = [r.doc.doc_id for r in docs]
        results = []
        queried_results = self.query(keys, *args, **kwargs)
        for d, r in zip(queried_results, docs):
            if d:
                r.doc.CopyFrom(d)
                _score = self.normalize_fn(r.score)
                _score = self.score_fn(_score, d)
                r.score.CopyFrom(_score)
            results.append(r)
        return results


class BaseKeyIndexer(BaseIndexer):

    def add(self, keys: List[Tuple[int, int]], weights: List[float], *args, **kwargs) -> int:
        pass

    def query(self, keys: List[int], *args, **kwargs) -> List[Tuple[int, int, float]]:
        pass


class JointIndexer(CompositionalTrainableBase):

    @property
    def components(self):
        return self._component

    @components.setter
    def components(self, comps: Callable[[], Union[list, dict]]):
        if not callable(comps):
            raise TypeError('component must be a callable function that returns '
                            'a List[BaseIndexer]')
        if not getattr(self, 'init_from_yaml', False):
            self._component = comps()
        else:
            self.logger.info('component is omitted from construction, '
                             'as it is initialized from yaml config')

        self._binary_indexer = None
        self._doc_indexer = None
        for c in self.components:
            if isinstance(c, BaseChunkIndexer):
                self._binary_indexer = c
            elif isinstance(c, BaseDocIndexer):
                self._doc_indexer = c
        if not self._binary_indexer or not self._doc_indexer:
            raise ValueError('"JointIndexer" requires a valid pair of "BaseBinaryIndexer" and "BaseTextIndexer"')

    def add(self, keys: Any, docs: Any, *args,
            **kwargs) -> None:
        if isinstance(docs, np.ndarray):
            self._binary_indexer.add(keys, docs, *args, **kwargs)
        elif isinstance(docs, list):
            self._doc_indexer.add(keys, docs, *args, **kwargs)
        else:
            raise TypeError('can not find an indexer for doc type: %s' % type(docs))

    def query(self,
              keys: Any,
              top_k: int,
              *args,
              **kwargs) -> List[List[Tuple]]:
        topk_results = self._binary_indexer.query(keys, top_k, *args, **kwargs)
        doc_caches = dict()
        topk_results_with_docs = []
        for topk in topk_results:
            topk_wd = []
            for doc_id, offset, weight, score in topk:
                doc = doc_caches.get(doc_id, self._doc_indexer.query([doc_id])[0])
                doc_caches[doc_id] = doc
                topk_wd.append((doc_id, offset, weight, score, doc.chunks[offset]))
            topk_results_with_docs.append(topk_wd)
        return topk_results_with_docs



import copy
import os
import unittest

from gnes.preprocessor.io_utils import ffmpeg
from gnes.preprocessor.io_utils import video
from gnes.preprocessor.io_utils import gif
from gnes.preprocessor.io_utils import audio


class TestFFmpeg(unittest.TestCase):
    def setUp(self):
        self.dirname = os.path.dirname(__file__)

        self.video_path = os.path.join(self.dirname, 'videos', 'test.mp4')
        self.frames = video.capture_frames(input_fn=self.video_path, fps=10, scale='640:360')

    def test_probe(self):
        probe = ffmpeg.probe(self.video_path)
        self.assertEqual(probe['height'], 720)
        self.assertEqual(probe['width'], 1280)
        self.assertEqual(probe['fps'], 25.0)


    def test_get_media_meta(self):
        meta1 = ffmpeg.get_media_meta(input_fn=self.video_path)
        with open(self.video_path, 'rb') as f:
            data = f.read()
            meta2 = ffmpeg.get_media_meta(input_data=data)
        self.assertEqual(meta1['frame_width'], meta2['frame_width'])
        self.assertEqual(meta1['frame_height'], meta2['frame_height'])

    def test_capture_frames(self):
        frames1 = video.capture_frames(input_fn=self.video_path, fps=10, scale='640:-2')

        with open(self.video_path, 'rb') as f:
            data = f.read()
            frames2 = video.capture_frames(input_data=data, fps=10, scale='-1:360')

        self.assertEqual(frames1.shape, frames2.shape)

    def test_scale_video(self):
        out = video.scale_video(input_fn=self.video_path, scale='640:360')
        meta = ffmpeg.get_media_meta(input_data=out, input_options={'format': 'mp4'})
        self.assertEqual(meta['frame_width'], 640)
        self.assertEqual(meta['frame_height'], 360)

    def test_encode_video(self):
        video_data = video.encode_video(images=self.frames)
        meta = ffmpeg.get_media_meta(input_data=video_data, input_options={'format': 'mp4'})
        self.assertEqual(meta['frame_width'], 640)
        self.assertEqual(meta['frame_height'], 360)

    def test_gif_encode(self):
        gif_data = gif.encode_video(images=self.frames, frame_rate=10)
        frames = gif.capture_frames(input_data=gif_data)
        self.assertEqual(self.frames.shape, frames.shape)

    def test_capture_audio(self):
        audio_data1 = audio.capture_audio(input_fn=self.video_path)
        with open(self.video_path, 'rb') as f:
            data = f.read()
            audio_data2 = audio.capture_audio(input_data=data)

            self.assertEqual(audio_data1.shape, audio_data2.shape)

    def test_split_audio(self):
        chunks1 = audio.split_audio(input_fn=self.video_path)
        with open(self.video_path, 'rb') as f:
            data = f.read()
            chunks2 = audio.split_audio(input_data=data)
            self.assertEqual(len(chunks1), len(chunks2))
            self.assertEqual(chunks1[0].shape, chunks2[0].shape)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the 'License');
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an 'AS IS' BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List
import numpy as np
import subprocess as sp
import tempfile

from .ffmpeg import compile_args, extract_frame_size
from .helper import _check_input, run_command


def capture_frames(input_fn: str = 'pipe:',
                   input_data: bytes = None,
                   fps: int = None,
                   pix_fmt: str = 'rgb24') -> 'np.ndarray':
    _check_input(input_fn, input_data)



    with tempfile.NamedTemporaryFile(suffix=".gif") as f:
        if input_data:
            f.write(input_data)
            f.flush()
            input_fn = f.name

        video_filters = []
        if fps:
            video_filters += ['fps=%d' % fps]

        output_kwargs = {'format': 'rawvideo', 'pix_fmt': pix_fmt}

        cmd_args = compile_args(
            input_fn=input_fn,
            video_filters=video_filters,
            output_options=output_kwargs)

        out, err = run_command(cmd_args, pipe_stdout=True, pipe_stderr=True)

        width, height = extract_frame_size(err.decode())

        depth = 3
        if pix_fmt == 'rgba':
            depth = 4

        frames = np.frombuffer(out,
                               np.uint8).reshape([-1, height, width, depth])
        return frames


def encode_video(images: 'np.ndarray', frame_rate: int, pix_fmt: str = 'rgb24'):

    cmd = [
        'ffmpeg', '-y', '-f', 'rawvideo', '-vcodec', 'rawvideo', '-r',
        '%.02f' % frame_rate, '-s',
        '%dx%d' % (images[0].shape[1], images[0].shape[0]), '-pix_fmt',
        'rgb24', '-i', '-', '-filter_complex',
        '[0:v]split[x][z];[z]palettegen[y];[x]fifo[x];[x][y]paletteuse', '-r',
        '%.02f' % frame_rate, '-f', 'gif', '-'
    ]
    proc = sp.Popen(cmd, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE)
    for image in images:
        proc.stdin.write(image.tostring())
    out, err = proc.communicate()
    if proc.returncode:
        err = '\n'.join([' '.join(cmd), err.decode('utf8')])
        raise IOError(err)
    del proc
    return out



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import grpc
import queue
from concurrent import futures

from gnes.proto import gnes_pb2_grpc


class BaseClient:

    def __init__(self, args):
        self.args = args

        self._channel = grpc.insecure_channel(
            '%s:%d' % (self.args.grpc_host, self.args.grpc_port),
            options={
                "grpc.max_send_message_length": -1,
                "grpc.max_receive_message_length": -1,
            }.items(),
        )

    def call(self, request):
        resp = self._stub.Call(request)
        return resp

    def async_call(self, request, callback_fn=None):
        response_future = self._stub.Call.future(self._request)
        if callback_fn:
            response_future.add_done_callback(callback_fn)
        else:
            return response_future

    def send_request(self, request):
        """Non-blocking wrapper for a client's request operation."""
        raise NotImplementedError

    def start(self):
        # waits for the channel to be ready before we start sending messages
        grpc.channel_ready_future(self._channel).result()
        self._stub = gnes_pb2_grpc.GnesRPCStub(self._channel)

    def stop(self):
        self._channel.close()
        self._stub = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class UnarySyncClient(BaseClient):

    def __init__(self, args):
        super().__init__(args)
        self._pool = futures.ThreadPoolExecutor(
            max_workers=self.args.max_concurrency)
        self.request_cnt = 0
        self._response_callbacks = []

    def send_request(self, request):
        # Send requests in seperate threads to support multiple outstanding rpcs
        self._pool.submit(self._dispatch_request, request)

    def stop(self):
        self._pool.shutdown(wait=True)
        super().stop()

    def _dispatch_request(self, request):
        resp = self._stub.Call(request)
        self._handle_response(self, resp)

    def _handle_response(self, client, response):
        for callback in self._response_callbacks:
            callback(client, response)

    def add_response_callback(self, callback):
        """callback will be invoked as callback(client, response)"""
        self._response_callbacks.append(callback)


class _SyncStream(object):

    def __init__(self, stub, handle_response):
        self._stub = stub
        self._handle_response = handle_response
        self._is_streaming = False
        self._request_queue = queue.Queue()

    def send_request(self, request):
        self._request_queue.put(request)

    def start(self):
        self._is_streaming = True
        response_stream = self._stub.StreamCall(self._request_generator())
        for resp in response_stream:
            self._handle_response(self, resp)

    def stop(self):
        self._is_streaming = False

    def _request_generator(self):
        while self._is_streaming:
            try:
                request = self._request_queue.get(block=True, timeout=1.0)
                yield request
            except queue.Empty:
                pass


class StreamingClient(UnarySyncClient):

    def __init__(self, args):
        super().__init__(args)

        self._streams = [
            _SyncStream(self._stub, self._handle_response)
            for _ in range(self.args.max_concurrency)
        ]
        self._curr_stream = 0

    def send_request(self, request):
        # Use a round_robin scheduler to determine what stream to send on
        self._streams[self._curr_stream].send_request(request)
        self._curr_stream = (self._curr_stream + 1) % len(self._streams)

    def start(self):
        super().start()
        for stream in self._streams:
            self._pool.submit(stream.start)

    def stop(self):
        for stream in self._streams:
            stream.stop()
        super().stop()


#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .base import get_unary_score, CombinedScoreFn


class WeightedChunkScoreFn(CombinedScoreFn):
    """score = d_chunk.weight * relevance * q_chunk.weight"""

    def __call__(self, last_score: 'gnes_pb2.Response.QueryResponse.ScoredResult.Score',
                 q_chunk: 'gnes_pb2.Chunk',
                 d_chunk: 'gnes_pb2.Chunk', *args, **kwargs):
        q_chunk_weight = get_unary_score(value=q_chunk.weight,
                                         name='query chunk weight',
                                         offset=q_chunk.offset)
        d_chunk_weight = get_unary_score(value=d_chunk.weight,
                                         name='document chunk weight',
                                         doc_id=d_chunk.doc_id,
                                         offset=d_chunk.offset)

        return super().__call__(last_score, q_chunk_weight, d_chunk_weight)

# TODO: write this as a class
#     @staticmethod
#     def eq2(q_chunk: 'gnes_pb2.Chunk', d_chunk: 'gnes_pb2.Chunk',
#             relevance, relevance_cls):
#         """
#         score = d_chunk.weight * relevance * offset_divergence * q_chunk.weight
#         offset_divergence is calculated based on doc_type:
#             TEXT && VIDEO && AUDIO: offset is 1-D
#             IMAGE: offset is 2-D
#         """
#
#         def _cal_divergence(q_chunk: 'gnes_pb2.Chunk', d_chunk: 'gnes_pb2.Chunk'):
#             if q_chunk.offset_nd and d_chunk.offset_nd:
#                 return 1 / (1 + np.sqrt((q_chunk.offset_nd[0] - d_chunk.offset_nd[0]) ** 2 +
#                                         (q_chunk.offset_nd[1] - d_chunk.offset_nd[1]) ** 2))
#             else:
#                 return np.abs(q_chunk.offset - d_chunk.offset)
#
#         score = gnes_pb2.Response.QueryResponse.ScoredResult.Score()
#
#         divergence = _cal_divergence(q_chunk, d_chunk)
#         score.value = d_chunk.weight * relevance * divergence * q_chunk.weight
#         score.explained = json.dumps({
#             'name': 'chunk_scorer_eq2',
#             'operand': [{'name': 'd_chunk_weight',
#                          'value': float(d_chunk.weight),
#                          'doc_id': d_chunk.doc_id,
#                          'offset': d_chunk.offset},
#                         {'name': 'q_chunk_weight',
#                          'value': float(q_chunk.weight),
#                          'offset': q_chunk.offset},
#                         {'name': 'relevance',
#                          'op': relevance_cls,
#                          'operand': [{'name': 'doc_chunk',
#                                       'doc_id': d_chunk.doc_id,
#                                       'offset': d_chunk.offset},
#                                      {'name': 'query_chunk',
#                                       'offset': q_chunk.offset}
#                                      ],
#                          'value': relevance
#                          },
#                         {'name': 'offset_divergence',
#                          'value': float(divergence)}],
#             'op': 'prod',
#             'value': float(score.value)
#         })
#         return score



import os
import unittest
import numpy as np

from gnes.encoder.base import BaseNumericEncoder


class TestQuantizerEncoder(unittest.TestCase):
    def setUp(self):
        dirname = os.path.dirname(__file__)
        self.vanilla_quantizer_yaml = os.path.join(dirname, 'yaml', 'quantizer_encoder.yml')

    def test_vanilla_quantizer(self):
        encoder = BaseNumericEncoder.load_yaml(self.vanilla_quantizer_yaml)
        encoder.train()

        vecs_1 = np.random.uniform(-150, 150, size=[1000, 160]).astype('float32')
        out = encoder.encode(vecs_1)
        self.assertEqual(len(out.shape), 2)
        self.assertEqual(out.shape[0], 1000)
        self.assertEqual(out.shape[1], 16)

        vecs_2 = np.random.uniform(-1, 1, size=[1000, 160]).astype('float32')
        self.assertRaises(Warning, encoder.encode, vecs_2)

        vecs_3 = np.random.uniform(-1, 1000, size=[1000, 160]).astype('float32')
        self.assertRaises(Warning, encoder.encode, vecs_3)




#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


# do not change this line
# this is managed by git tag and replaced on every release
__version__ = '0.0.37'
__proto_version__ = '20190905'



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from collections import defaultdict
from typing import List, Generator

from gnes.score_fn.base import CombinedScoreFn
from ..base import TrainableBase, CompositionalTrainableBase
from ..proto import gnes_pb2, merge_routes


class BaseRouter(TrainableBase):
    """ Base class for the router. Inherit from this class to create a new router.

    Router forwards messages between services. Essentially, it receives a 'gnes_pb2.Message'
    and call `apply()` method on it.
    """

    def apply(self, msg: 'gnes_pb2.Message', *args, **kwargs):
        """
        Modify the incoming message

        :param msg: incoming message
        """
        pass


class BaseMapRouter(BaseRouter):
    def apply(self, msg: 'gnes_pb2.Message', *args, **kwargs) -> Generator:
        pass


class BaseReduceRouter(BaseRouter):
    def apply(self, msg: 'gnes_pb2.Message', accum_msgs: List['gnes_pb2.Message'], *args, **kwargs) -> None:
        """
        Modify the current message based on accumulated messages

        :param msg: the current message
        :param accum_msgs: accumulated messages
        """
        merge_routes(msg, accum_msgs)
        if len(msg.envelope.num_part) > 1:
            msg.envelope.num_part.pop()
        else:
            self.logger.warning(
                'message envelope says num_part=%s, means no further message reducing. '
                'ignore this if you explicitly set "num_part" in RouterService' % msg.envelope.num_part)


class BaseTopkReduceRouter(BaseReduceRouter):
    def __init__(self, reduce_op: str = 'sum', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._reduce_op = reduce_op

    def post_init(self):
        self.reduce_op = CombinedScoreFn(score_mode=self._reduce_op)

    def get_key(self, x: 'gnes_pb2.Response.QueryResponse.ScoredResult') -> str:
        raise NotImplementedError

    def set_key(self, x: 'gnes_pb2.Response.QueryResponse.ScoredResult', k: str) -> None:
        raise NotImplementedError

    def apply(self, msg: 'gnes_pb2.Message', accum_msgs: List['gnes_pb2.Message'], *args, **kwargs):
        # now convert chunk results to doc results
        all_scored_results = [sr for m in accum_msgs for sr in m.response.search.topk_results]
        score_dict = defaultdict(list)

        # count score by iterating over chunks
        for c in all_scored_results:
            score_dict[self.get_key(c)].append(c.score)

        for k, v in score_dict.items():
            score_dict[k] = self.reduce_op(*v)

        msg.response.search.ClearField('topk_results')

        for k, v in score_dict.items():
            r = msg.response.search.topk_results.add()
            r.score.CopyFrom(v)
            self.set_key(r, k)

        super().apply(msg, accum_msgs)


class PipelineRouter(CompositionalTrainableBase):
    def apply(self, *args, **kwargs) -> None:
        if not self.components:
            raise NotImplementedError
        for be in self.components:
            be.apply(*args, **kwargs)



import os
import unittest
from shutil import rmtree

from gnes.indexer.doc.dict import DictIndexer
from gnes.indexer.doc.leveldb import LVDBIndexer
from gnes.proto import gnes_pb2
from tests import txt_file2pb_docs


class TestBaseLVDB(unittest.TestCase):
    def setUp(self):
        dirname = os.path.dirname(__file__)

        self.test_docs = txt_file2pb_docs(open(os.path.join(dirname, 'tangshi.txt'), encoding='utf8'))

        self.db_path = './test_leveldb'
        self.dump_path = os.path.join(dirname, 'indexer.bin')
        self.dump_yaml_path = os.path.join(dirname, 'indexer.yaml')

    def tearDown(self):
        if os.path.exists(self.db_path):
            rmtree(self.db_path)
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)
        if os.path.exists(self.dump_yaml_path):
            os.remove(self.dump_yaml_path)
        if os.path.exists('my-indexer-531.bin'):
            os.remove('my-indexer-531.bin')
        if os.path.exists('my-indexer-531.yml'):
            os.remove('my-indexer-531.yml')

    def test_dict_indexer(self):
        db = DictIndexer()
        db.add(range(len(self.test_docs)), self.test_docs)
        db.dump(self.dump_path)
        self.assertEqual(len(self.test_docs), db.size)
        db2 = DictIndexer.load(self.dump_path)
        self.assertEqual(len(self.test_docs), db2.size)
        db.name = 'my-indexer-531'
        db.dump()
        db.dump_yaml()
        db3 = DictIndexer.load_yaml(db.yaml_full_path)
        for k in db3.query([1, 2, 3]):
            self.assertIsInstance(k, gnes_pb2.Document)
        self.assertEqual(len(self.test_docs), db3.size)

    def test_add_docs(self):
        db = LVDBIndexer(self.db_path)
        db.add(range(len(self.test_docs)), self.test_docs)
        self.assertTrue(os.path.exists(self.db_path))
        self.assertLess(0, len(os.listdir(self.db_path)))
        db.close()

    def test_query(self):
        db = LVDBIndexer(self.db_path)
        db.add(range(len(self.test_docs)), self.test_docs)

        res1 = db.query(range(len(self.test_docs)))
        num_non_empty = sum(1 for d in res1 if d)
        self.assertEqual(num_non_empty, len(self.test_docs))

        res2 = db.query(range(len(self.test_docs) + 1, len(self.test_docs) + 100))
        num_non_empty = sum(1 for d in res2 if d)
        self.assertEqual(num_non_empty, 0)
        db.close()

    def dump_load(self):
        tmp = LVDBIndexer(self.db_path)
        tmp.add(range(len(self.test_docs)), self.test_docs)
        tmp.dump(self.dump_path)
        tmp.close()

        db = LVDBIndexer.load(self.db_path)
        res1 = db.query(range(len(self.test_docs)))
        num_non_empty = sum(1 for d in res1 if d)
        self.assertEqual(num_non_empty, self.test_data1.length)

        res2 = db.query(range(len(self.test_docs) + 1, len(self.test_docs) + 100))
        num_non_empty = sum(1 for d in res2 if d)
        self.assertEqual(num_non_empty, 0)
        db.close()



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import List, Tuple, Any

import numpy as np

from ..base import BaseChunkIndexer
from ..key_only import ListKeyIndexer


class NumpyIndexer(BaseChunkIndexer):
    """An exhaustive search indexer using numpy
    The distance is computed as L1 distance normalized by the number of dimension
    """

    def __init__(self, is_binary: bool = False, *args, **kwargs):
        super().__init__()
        self._num_dim = None
        self._vectors = None  # type: np.ndarray
        self._is_binary = is_binary
        self._key_info_indexer = ListKeyIndexer()

    def add(self, keys: List[Tuple[int, Any]], vectors: np.ndarray, weights: List[float], *args,
            **kwargs):
        if len(vectors) % len(keys) != 0:
            raise ValueError('vectors bytes should be divided by doc_ids')

        if not self._num_dim:
            self._num_dim = vectors.shape[1]
        elif self._num_dim != vectors.shape[1]:
            raise ValueError(
                "vectors' shape [%d, %d] does not match with indexer's dim: %d" %
                (vectors.shape[0], vectors.shape[1], self._num_dim))

        if self._vectors is not None:
            self._vectors = np.concatenate([self._vectors, vectors], axis=0)
        else:
            self._vectors = vectors
        self._key_info_indexer.add(keys, weights)

    def query(self, keys: np.ndarray, top_k: int, *args, **kwargs) -> List[List[Tuple]]:
        dist = np.abs(np.expand_dims(keys, axis=1) - np.expand_dims(self._vectors, axis=0))

        if self._is_binary:
            dist = np.minimum(dist, 1)

        score = np.sum(dist, -1) / self._num_dim

        ret = []
        for ids in score:
            rk = sorted(enumerate(ids), key=lambda x: x[1])[:top_k]
            chunk_info = self._key_info_indexer.query([j[0] for j in rk])
            ret.append([(*r, s) for r, s in zip(chunk_info, [j[1] for j in rk])])
        return ret



import os
import unittest
import numpy as np

from gnes.encoder.base import BaseEncoder


class TestYT8MFeatureExtractor(unittest.TestCase):
    @unittest.skip
    def setUp(self):
        dirname = os.path.dirname(__file__)
        self.test_img = [np.random.rand(10, 299, 299, 3).astype(np.uint8),
                         np.random.rand(6, 299, 299, 3).astype(np.uint8),
                         np.random.rand(8, 299, 299, 3).astype(np.uint8)]
        self.inception_yaml = os.path.join(dirname, 'yaml', 'yt8m_feature.yml')

    @unittest.skip
    def test_inception_encoding(self):
        self.encoder = BaseEncoder.load_yaml(self.inception_yaml)
        vec = self.encoder.encode(self.test_img)

        self.assertEqual(len(vec), len(self.test_img))
        self.assertEqual(len(vec[0].shape), 2)
        self.assertEqual(vec[0].shape[0], self.test_img[0].shape[0])
        self.assertEqual(vec[0].shape[1], 1152)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import sys
import time
import zipfile
from math import ceil
from typing import List

from termcolor import colored

from ..proto import RequestGenerator
from .grpc import StreamingSyncClient


class CLIClient(StreamingSyncClient):
    def __init__(self, args):
        super().__init__(args)
        self.start()
        getattr(self, self.args.mode)(all_bytes)
        self.stop()

    def train(self, all_bytes: List[bytes], stub):
        with ProgressBar(all_bytes, self.args.batch_size, task_name=self.args.mode) as p_bar:
            for req in RequestGenerator.train(all_bytes, doc_id_start=self.args.start_doc_id, batch_size=self.args.batch_size):
                self.send_request(req)
                p_bar.update()

    def index(self, all_bytes: List[bytes], stub):
        with ProgressBar(all_bytes, self.args.batch_size, task_name=self.args.mode) as p_bar:
            for req in RequestGenerator.index(all_bytes, doc_id_start=self.args.start_doc_id, batch_size=self.args.batch_size):
                self.send_request(req)
                p_bar.update()

    def query(self, all_bytes: List[bytes], stub):
        for idx, q in enumerate(all_bytes):
            for req in RequestGenerator.query(q, request_id_start=idx, top_k=self.args.top_k):
                resp = self._stub.Call(req)
                print(resp)
                print('query %d result: %s' % (idx, resp))
                input('press any key to continue...')

    def read_all(self) -> List[bytes]:
        if self.args.txt_file:
            all_bytes = [v.encode() for v in self.args.txt_file]
        elif self.args.image_zip_file:
            zipfile_ = zipfile.ZipFile(self.args.image_zip_file)
            all_bytes = [zipfile_.open(v).read() for v in zipfile_.namelist()]
        elif self.args.video_zip_file:
            zipfile_ = zipfile.ZipFile(self.args.video_zip_file)
            all_bytes = [zipfile_.open(v).read() for v in zipfile_.namelist()]
        else:
            raise AttributeError('--txt_file, --image_zip_file, --video_zip_file one must be given')

        return all_bytes


class ProgressBar:
    def __init__(self, all_bytes: List[bytes], batch_size: int, bar_len: int = 20, task_name: str = ''):
        self.all_bytes_len = [len(v) for v in all_bytes]
        self.batch_size = batch_size
        self.total_batch = ceil(len(self.all_bytes_len) / self.batch_size)
        self.bar_len = bar_len
        self.task_name = task_name

    def update(self):
        if self.num_batch > self.total_batch - 1:
            return
        sys.stdout.write('\r')
        elapsed = time.perf_counter() - self.start_time
        elapsed_str = colored('elapsed', 'yellow')
        speed_str = colored('speed', 'yellow')
        estleft_str = colored('left', 'yellow')
        self.num_batch += 1
        percent = self.num_batch / self.total_batch
        num_bytes = sum(self.all_bytes_len[((self.num_batch - 1) * self.batch_size):(self.num_batch * self.batch_size)])
        sys.stdout.write(
            '{:>10} [{:<{}}] {:3.0f}%   {:>8}: {:3.1f}s   {:>8}: {:3.1f} bytes/s  {:3.1f} batch/s {:>8}: {:3.1f}s'.format(
                colored(self.task_name, 'cyan'),
                colored('=' * int(self.bar_len * percent), 'green'),
                self.bar_len + 9,
                percent * 100,
                elapsed_str,
                elapsed,
                speed_str,
                num_bytes / elapsed,
                self.num_batch / elapsed,
                estleft_str,
                (self.total_batch - self.num_batch) / ((self.num_batch + 0.0001) / elapsed)
            ))
        sys.stdout.flush()

    def __enter__(self):
        self.start_time = time.perf_counter()
        self.num_batch = -1
        sys.stdout.write('\n')
        self.update()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.write('\t%s\n' % colored('done!', 'green'))



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import grpc

from ..helper import TimeContext
from ..proto import gnes_pb2_grpc, RequestGenerator


class BenchmarkClient:
    def __init__(self, args):

        all_bytes = [b'a' * args.request_length] * args.num_requests

        with grpc.insecure_channel(
                '%s:%d' % (args.grpc_host, args.grpc_port),
                options=[('grpc.max_send_message_length', args.max_message_size * 1024 * 1024),
                         ('grpc.max_receive_message_length', args.max_message_size * 1024 * 1024)]) as channel:
            stub = gnes_pb2_grpc.GnesRPCStub(channel)

            id = 0
            with TimeContext('StreamCall') as tc:
                resp = stub.StreamCall(RequestGenerator.index(all_bytes, args.batch_size))
                for r in resp:
                    assert r.request_id == id
                    id += 1
            stream_call_el = tc.duration

            with TimeContext('Call') as tc:
                for req in RequestGenerator.index(all_bytes, batch_size=1):
                    r = stub.Call(req)
                    assert r.request_id == req.request_id

            call_el = tc.duration

        print('num_requests     %d\n'
              'request_length   %d' % (args.num_requests, args.request_length))
        print('StreamCall       %3.3f s\n'
              'Call             %3.3f s\n' % (stream_call_el, call_el))
